from typing import Dict, List, Iterable, Tuple
from multiprocessing import Pool, get_context
from pathlib import Path

from gensim.similarities.annoy import AnnoyIndexer
from gensim.similarities import SoftCosineSimilarity, LevenshteinSimilarityIndex, WordEmbeddingSimilarityIndex, SparseTermSimilarityMatrix
from gensim.corpora import Dictionary
from gensim.utils import simple_preprocess as preprocess
from gensim.models import TfidfModel
from more_itertools import chunked, zip_equal
import numpy as np

from .document import Document
from .fasttext import FastText
from tqdm.autonotebook import tqdm


DICTIONARY = None


def get_document_text(document: Document) -> List[str]:
    return preprocess(document.text)


def get_document_texts(corpus: List[Document], desc: str) -> Iterable[List[str]]:
    with Pool(None) as pool:
        texts = pool.imap(get_document_text, corpus)
        for text in tqdm(texts, total=len(corpus), desc=desc):
            yield text


def doc2bow(document: str) -> List[Tuple[int, int]]:
    return DICTIONARY.doc2bow(document)


def get_results(corpus: List[Document], similarities: List[float], topn: int = 1000) -> List[Tuple[Document, float]]:
    results = sorted(zip(corpus, similarities), key=lambda x: (-x[1], x[0]))[:topn]
    return results


class Index:
    BASE_PATH = Path('cache')

    def __init__(self, documents: Dict[str, Document]):
        self.corpus = sorted(documents.values())

        dictionary_filename = str(self.BASE_PATH / 'dictionary')
        try:
            print('Loading {}'.format(dictionary_filename))
            self.dictionary = Dictionary.load(dictionary_filename)
        except IOError:
            print('Creating {}'.format(dictionary_filename))
            self.dictionary = Dictionary(get_document_texts(self.corpus, 'Loading document texts'))
            self.dictionary.save(dictionary_filename)

        self.query_tfidf = TfidfModel(dictionary=self.dictionary, smartirs='ltb', slope=0.2)

        index_filename = str(self.BASE_PATH / 'SCM-index')
        try:
            print('Loading {}'.format(index_filename))
            self.index = SoftCosineSimilarity.load(index_filename)
        except IOError:
            global DICTIONARY
            DICTIONARY = self.dictionary
            # Default context on some OSs ("spawn") does not allow to access the shared objects (DICTIONARY)
            with get_context('fork').Pool(None) as pool:
                bow_corpus = list(pool.imap(doc2bow, get_document_texts(self.corpus, 'Creating BoW corpus')))
            DICTIONARY = None
            document_tfidf = TfidfModel(dictionary=self.dictionary, smartirs='lnb', slope=0.2)
            tfidf_corpus = document_tfidf[bow_corpus]

            we_similarity_matrix_filename = str(self.BASE_PATH / 'SCM-WE-pretrained_similarity-matrix')
            try:
                print('Loading {}'.format(we_similarity_matrix_filename))
                we_similarity_matrix = SparseTermSimilarityMatrix.load(we_similarity_matrix_filename)
            except IOError:
                print('Creating {}'.format(we_similarity_matrix_filename))
                word_embeddings = FastText().wv
                indexer = AnnoyIndexer(word_embeddings, num_trees=1)
                we_word_similarities = WordEmbeddingSimilarityIndex(word_embeddings, threshold=-1.0, exponent=4.0, kwargs={'indexer': indexer})
                we_similarity_matrix = SparseTermSimilarityMatrix(we_word_similarities, self.dictionary, document_tfidf, symmetric=True, dominant=True, nonzero_limit=100)
                we_similarity_matrix.save(we_similarity_matrix_filename)

            ld_similarity_matrix_filename = str(self.BASE_PATH / 'SCM-LD_similarity-matrix')
            try:
                print('Loading {}'.format(ld_similarity_matrix_filename))
                ld_similarity_matrix = SparseTermSimilarityMatrix.load(ld_similarity_matrix_filename)
            except IOError:
                ld_word_similarities_filename = str(self.BASE_PATH / 'SCM-LD_similarity-index')
                try:
                    print('Loading {}'.format(ld_word_similarities_filename))
                    ld_word_similarities = LevenshteinSimilarityIndex.load(ld_word_similarities_filename)
                except IOError:
                    print('Creating {}'.format(ld_word_similarities_filename))
                    ld_word_similarities = LevenshteinSimilarityIndex(self.dictionary, alpha=1.8, beta=5.0, max_distance=3)
                    ld_word_similarities.save(ld_word_similarities_filename)
                print('Creating {}'.format(ld_similarity_matrix_filename))
                ld_similarity_matrix = SparseTermSimilarityMatrix(ld_word_similarities, self.dictionary, document_tfidf, symmetric=True, dominant=False, nonzero_limit=100)
                ld_similarity_matrix.save(ld_similarity_matrix_filename)
            combined_similarity_matrix = SparseTermSimilarityMatrix((0.1 * we_similarity_matrix.matrix + 0.9 * ld_similarity_matrix.matrix))

            print('Creating {}'.format(index_filename))
            self.index = SoftCosineSimilarity(tfidf_corpus, combined_similarity_matrix, normalized=('maintain', 'maintain'))
            self.index.save(index_filename)

    def bulk_query(self, queries: Iterable[str], chunk_size: int = 1000) -> Iterable[Tuple[str, Tuple[Document, float]]]:
        queries = list(queries)
        progress_bar = iter(tqdm(queries, desc='Querying the index'))
        for query_chunk in chunked(queries, chunk_size):
            global DICTIONARY
            DICTIONARY = self.dictionary
            # Default context on some OSs ("spawn") does not allow to access the shared objects (DICTIONARY)
            with get_context('fork').Pool(None) as pool:
                bow_query_chunk = list(pool.imap(doc2bow, pool.imap(preprocess, query_chunk)))
            DICTIONARY = None
            tfidf_query_chunk = self.query_tfidf[bow_query_chunk]
            with np.errstate(divide='ignore', invalid='ignore'):
                similarity_chunk = self.index[tfidf_query_chunk]
            for query, similarities in zip_equal(query_chunk, similarity_chunk):
                results = get_results(self.corpus, similarities)
                yield query, results
                next(progress_bar)
